@@ -12,10 +12,16 @@
 -----------------------------------------------------------------------------
 
 {-# LANGUAGE CPP                 #-}
+{-# LANGUAGE DataKinds           #-}
+{-# LANGUAGE DeriveAnyClass      #-}
+{-# LANGUAGE DeriveDataTypeable  #-}
 {-# LANGUAGE FlexibleContexts    #-}
 {-# LANGUAGE Rank2Types          #-}
+{-# LANGUAGE QuasiQuotes         #-}
 {-# LANGUAGE ScopedTypeVariables #-}
-{-# LANGUAGE DataKinds           #-}
+{-# LANGUAGE StandaloneDeriving  #-}
+{-# LANGUAGE TemplateHaskell     #-}
+{-# LANGUAGE TypeApplications    #-}
 
 #if MIN_VERSION_base(4,19,0)
 {-# OPTIONS_GHC -Wall -Werror -Wno-incomplete-uni-patterns -Wno-x-partial #-}
@@ -25,16 +31,20 @@
 
 module TestSuite.Basics.ArithSolver(tests) where
 
-import Data.SBV.Internals hiding (free, free_)
+import Data.SBV.Internals hiding (free, free_, (#))
 import Utils.SBVTestFramework
 
 import Data.List (genericIndex, isInfixOf, isPrefixOf, isSuffixOf, genericTake, genericDrop, genericLength)
 
 import qualified Data.Char       as C
 import qualified Data.SBV.Char   as SC
-import qualified Data.SBV.String as SS
 import qualified Data.SBV.List   as SL
 
+import Data.SBV.Rational
+
+data Day = Mon | Tue | Wed | Thu | Fri | Sat | Sun deriving (Bounded, Enum, Eq)
+mkSymbolicEnumeration  ''Day
+
 -- Test suite
 tests :: TestTree
 tests =
@@ -82,6 +92,8 @@ tests =
      ++ genChars
      ++ genStrings
      ++ genLists
+     ++ genEnums
+     ++ misc
      )
 
 genExtends :: [TestTree]
@@ -140,6 +152,8 @@ genBoolTest nm op opS = map mkTest $  [(show x, show y, mkThm2  x y (x `op` y))
                                    ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- i64s,      y <- i64s     ]
                                    ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- iUBs,      y <- iUBs     ]
                                    ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- reducedCS, y <- reducedCS]
+                                   ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- fs,        y <- fs       ]
+                                   ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- ds,        y <- ds       ]
                                    ++ [(show x, show y, mkThm2  x y (x `op` y)) |                             x <- ss,        y <- ss       ]
                                    ++ [(show x, show y, mkThm2L x y (x `op` y)) | nm `elem` allowedListComps, x <- sl,        y <- sl       ]
                                    ++ [(show x, show y, mkThm2M x y (x `op` y)) |                             x <- sm,        y <- sm       ]
@@ -327,16 +341,16 @@ genIntCasts = map mkTest $  cast w8s ++ cast w16s ++ cast w32s ++ cast w64s
                                       return $ literal res .== sFromIntegral a
 
 genReals :: [TestTree]
-genReals = map mkTest $  [("+",  show x, show y, mkThm2 (+)   x y (x +  y)) | x <- rs, y <- rs        ]
-                      ++ [("-",  show x, show y, mkThm2 (-)   x y (x -  y)) | x <- rs, y <- rs        ]
-                      ++ [("*",  show x, show y, mkThm2 (*)   x y (x *  y)) | x <- rs, y <- rs        ]
-                      ++ [("/",  show x, show y, mkThm2 (/)   x y (x /  y)) | x <- rs, y <- rs, y /= 0]
-                      ++ [("<",  show x, show y, mkThm2 (.<)  x y (x <  y)) | x <- rs, y <- rs        ]
-                      ++ [("<=", show x, show y, mkThm2 (.<=) x y (x <= y)) | x <- rs, y <- rs        ]
-                      ++ [(">",  show x, show y, mkThm2 (.>)  x y (x >  y)) | x <- rs, y <- rs        ]
-                      ++ [(">=", show x, show y, mkThm2 (.>=) x y (x >= y)) | x <- rs, y <- rs        ]
-                      ++ [("==", show x, show y, mkThm2 (.==) x y (x == y)) | x <- rs, y <- rs        ]
-                      ++ [("/=", show x, show y, mkThm2 (./=) x y (x /= y)) | x <- rs, y <- rs        ]
+genReals = map mkTest $  [("+",  show x, show y, mkThm2 (+)   x y (x +  y)) | x <- ars, y <- ars        ]
+                      ++ [("-",  show x, show y, mkThm2 (-)   x y (x -  y)) | x <- ars, y <- ars        ]
+                      ++ [("*",  show x, show y, mkThm2 (*)   x y (x *  y)) | x <- ars, y <- ars        ]
+                      ++ [("/",  show x, show y, mkThm2 (/)   x y (x /  y)) | x <- ars, y <- ars, y /= 0]
+                      ++ [("<",  show x, show y, mkThm2 (.<)  x y (x <  y)) | x <- ars, y <- ars        ]
+                      ++ [("<=", show x, show y, mkThm2 (.<=) x y (x <= y)) | x <- ars, y <- ars        ]
+                      ++ [(">",  show x, show y, mkThm2 (.>)  x y (x >  y)) | x <- ars, y <- ars        ]
+                      ++ [(">=", show x, show y, mkThm2 (.>=) x y (x >= y)) | x <- ars, y <- ars        ]
+                      ++ [("==", show x, show y, mkThm2 (.==) x y (x == y)) | x <- ars, y <- ars        ]
+                      ++ [("/=", show x, show y, mkThm2 (./=) x y (x /= y)) | x <- ars, y <- ars        ]
   where mkTest (nm, x, y, t) = testCase ("genReals.arithmetic-" ++ nm ++ "." ++ x ++ "_" ++ y) (assert t)
         mkThm2 op x y r = isTheorem $ do [a, b] <- mapM free ["x", "y"]
                                          constrain $ a .== literal x
@@ -349,7 +363,7 @@ genFloats = genIEEE754 "genFloats" fs
 genDoubles :: [TestTree]
 genDoubles = genIEEE754 "genDoubles" ds
 
-genIEEE754 :: (IEEEFloating a, Show a) => String -> [a] -> [TestTree]
+genIEEE754 :: (IEEEFloating a, Num (SBV a), Show a) => String -> [a] -> [TestTree]
 genIEEE754 origin vs =  [tst1 ("pred_"   ++ nm, x, y)    | (nm, x, y)    <- preds]
                      ++ [tst1 ("unary_"  ++ nm, x, y)    | (nm, x, y)    <- uns]
                      ++ [tst2 ("binary_" ++ nm, x, y, r) | (nm, x, y, r) <- bins]
@@ -460,7 +474,7 @@ genFPConverts = [tst1 ("fpCast_" ++ nm, x, y) | (nm, x, y) <- converts]
                  ++  [("toFP_Float_ToFloat",    show x, mkThm1 (m toSFloat) x                           x  ) | x <- fs  ]
                  ++  [("toFP_Double_ToFloat",   show x, mkThm1 (m toSFloat) x (                   fp2fp x )) | x <- ds  ]
                  ++  [("toFP_Integer_ToFloat",  show x, mkThmC (m toSFloat) x (fromRational (toRational x))) | x <- iUBs]
-                 ++  [("toFP_Real_ToFloat",     show x, mkThmC (m toSFloat) x (fromRational (toRational x))) | x <- rs  ]
+                 ++  [("toFP_Real_ToFloat",     show x, mkThmC (m toSFloat) x (fromRational (toRational x))) | x <- ars ]
 
                  ++  [("toFP_Int8_ToDouble",    show x, mkThmC (m toSDouble) x (fromRational (toRational x))) | x <- i8s ]
                  ++  [("toFP_Int16_ToDouble",   show x, mkThmC (m toSDouble) x (fromRational (toRational x))) | x <- i16s]
@@ -473,7 +487,7 @@ genFPConverts = [tst1 ("fpCast_" ++ nm, x, y) | (nm, x, y) <- converts]
                  ++  [("toFP_Float_ToDouble",   show x, mkThm1 (m toSDouble) x (                   fp2fp x )) | x <- fs  ]
                  ++  [("toFP_Double_ToDouble",  show x, mkThm1 (m toSDouble) x                           x )  | x <- ds  ]
                  ++  [("toFP_Integer_ToDouble", show x, mkThmC (m toSDouble) x (fromRational (toRational x))) | x <- iUBs]
-                 ++  [("toFP_Real_ToDouble",    show x, mkThmC (m toSDouble) x (fromRational (toRational x))) | x <- rs  ]
+                 ++  [("toFP_Real_ToDouble",    show x, mkThmC (m toSDouble) x (fromRational (toRational x))) | x <- ars ]
 
                  -- Conversions from floats are only well-defined if the input is in-bounds. So we just check round-trip for these.
                  -- Also note that we clamp Int32/Word32/Int64/Word64 conversions further as floats become too sparse to handle those.
@@ -575,32 +589,39 @@ genQRems = map mkTest $  [("divMod",  show x, show y, mkThm2 sDivMod  x y (x `di
         noOverflow x y = not (x == minBound && y == -1)
 
 genChars :: [TestTree]
-genChars = map mkTest $  [("ord",           show c, mkThm SC.ord             cord            c) | c <- cs]
-                      ++ [("toLower",       show c, mkThm SC.toLowerL1       C.toLower       c) | c <- cs]
-                      ++ [("toUpper",       show c, mkThm SC.toUpperL1       C.toUpper       c) | c <- cs, toUpperExceptions c]
-                      ++ [("digitToInt",    show c, mkThm SC.digitToInt      dig2Int         c) | c <- cs, digitToIntRange c]
-                      ++ [("intToDigit",    show c, mkThm SC.intToDigit      int2Dig         c) | c <- [0 .. 15]]
-                      ++ [("isControl",     show c, mkThm SC.isControlL1     C.isControl     c) | c <- cs]
-                      ++ [("isSpace",       show c, mkThm SC.isSpaceL1       C.isSpace       c) | c <- cs]
-                      ++ [("isLower",       show c, mkThm SC.isLowerL1       C.isLower       c) | c <- cs]
-                      ++ [("isUpper",       show c, mkThm SC.isUpperL1       C.isUpper       c) | c <- cs]
-                      ++ [("isAlpha",       show c, mkThm SC.isAlphaL1       C.isAlpha       c) | c <- cs]
-                      ++ [("isAlphaNum",    show c, mkThm SC.isAlphaNumL1    C.isAlphaNum    c) | c <- cs]
-                      ++ [("isPrint",       show c, mkThm SC.isPrintL1       C.isPrint       c) | c <- cs]
-                      ++ [("isDigit",       show c, mkThm SC.isDigit         C.isDigit       c) | c <- cs]
-                      ++ [("isOctDigit",    show c, mkThm SC.isOctDigit      C.isOctDigit    c) | c <- cs]
-                      ++ [("isHexDigit",    show c, mkThm SC.isHexDigit      C.isHexDigit    c) | c <- cs]
-                      ++ [("isLetter",      show c, mkThm SC.isLetterL1      C.isLetter      c) | c <- cs]
-                      ++ [("isMark",        show c, mkThm SC.isMarkL1        C.isMark        c) | c <- cs]
-                      ++ [("isNumber",      show c, mkThm SC.isNumberL1      C.isNumber      c) | c <- cs]
-                      ++ [("isPunctuation", show c, mkThm SC.isPunctuationL1 C.isPunctuation c) | c <- cs]
-                      ++ [("isSymbol",      show c, mkThm SC.isSymbolL1      C.isSymbol      c) | c <- cs]
-                      ++ [("isSeparator",   show c, mkThm SC.isSeparatorL1   C.isSeparator   c) | c <- cs]
-                      ++ [("isAscii",       show c, mkThm SC.isAscii         C.isAscii       c) | c <- cs]
-                      ++ [("isLatin1",      show c, mkThm SC.isLatin1        C.isLatin1      c) | c <- cs]
-                      ++ [("isAsciiUpper",  show c, mkThm SC.isAsciiUpper    C.isAsciiUpper  c) | c <- cs]
-                      ++ [("isAsciiLower",  show c, mkThm SC.isAsciiLower    C.isAsciiLower  c) | c <- cs]
-  where toUpperExceptions = (`notElem` "\181\255")
+genChars = [ testCase "solver_genChars" (assert (isTheorem t)) ]
+  where t = do a <- free "a"
+               i <- free "i"
+
+               let chk  sop cop v = (a .== literal v) .=> sop a .== literal (cop v)
+                   chkI sop cop v = (i .== literal v) .=> sop i .== literal (cop v)
+
+               pure $ sAnd $  [chk  SC.ord             cord            c | c <- cs]
+                           ++ [chk  SC.toLowerL1       C.toLower       c | c <- cs]
+                           ++ [chk  SC.toUpperL1       C.toUpper       c | c <- cs]
+                           ++ [chk  SC.digitToInt      dig2Int         c | c <- cs, digitToIntRange c]
+                           ++ [chkI SC.intToDigit      int2Dig         c | c <- [0 .. 15]]
+                           ++ [chk  SC.isControlL1     C.isControl     c | c <- cs]
+                           ++ [chk  SC.isSpaceL1       C.isSpace       c | c <- cs]
+                           ++ [chk  SC.isLowerL1       C.isLower       c | c <- cs]
+                           ++ [chk  SC.isUpperL1       C.isUpper       c | c <- cs]
+                           ++ [chk  SC.isAlphaL1       C.isAlpha       c | c <- cs]
+                           ++ [chk  SC.isAlphaNumL1    C.isAlphaNum    c | c <- cs]
+                           ++ [chk  SC.isPrintL1       C.isPrint       c | c <- cs]
+                           ++ [chk  SC.isDigit         C.isDigit       c | c <- cs]
+                           ++ [chk  SC.isOctDigit      C.isOctDigit    c | c <- cs]
+                           ++ [chk  SC.isHexDigit      C.isHexDigit    c | c <- cs]
+                           ++ [chk  SC.isLetterL1      C.isLetter      c | c <- cs]
+                           ++ [chk  SC.isMarkL1        C.isMark        c | c <- cs]
+                           ++ [chk  SC.isNumberL1      C.isNumber      c | c <- cs]
+                           ++ [chk  SC.isPunctuationL1 C.isPunctuation c | c <- cs]
+                           ++ [chk  SC.isSymbolL1      C.isSymbol      c | c <- cs]
+                           ++ [chk  SC.isSeparatorL1   C.isSeparator   c | c <- cs]
+                           ++ [chk  SC.isAscii         C.isAscii       c | c <- cs]
+                           ++ [chk  SC.isLatin1        C.isLatin1      c | c <- cs]
+                           ++ [chk  SC.isAsciiUpper    C.isAsciiUpper  c | c <- cs]
+                           ++ [chk  SC.isAsciiLower    C.isAsciiLower  c | c <- cs]
+
         digitToIntRange   = (`elem` "0123456789abcdefABCDEF")
         cord :: Char -> Integer
         cord = fromIntegral . C.ord
@@ -608,32 +629,27 @@ genChars = map mkTest $  [("ord",           show c, mkThm SC.ord             cor
         dig2Int = fromIntegral . C.digitToInt
         int2Dig :: Integer -> Char
         int2Dig = C.intToDigit . fromIntegral
-        mkTest (nm, x, t) = testCase ("genChars-" ++ nm ++ "." ++ x) (assert t)
-        mkThm sop cop arg = isTheorem $ do a <- free "a"
-                                           constrain $ a .== literal arg
-                                           return $ literal (cop arg) .== sop a
 
 genStrings :: [TestTree]
-genStrings = map mkTest1 (  [("length",        show s,                   mkThm1 SS.length        strLen        s      ) | s <- ss                                                       ]
-                         ++ [("null",          show s,                   mkThm1 SS.null          null          s      ) | s <- ss                                                       ]
-                         ++ [("head",          show s,                   mkThm1 SS.head          head          s      ) | s <- ss, not (null s)                                         ]
-                         ++ [("tail",          show s,                   mkThm1 SS.tail          tail          s      ) | s <- ss, not (null s)                                         ]
-                         ++ [("singleton",     show c,                   mkThm1 SS.singleton     (: [])        c      ) | c <- cs                                                       ]
-                         ++ [("implode",       show s,                   mkThmI SS.implode                     s      ) | s <- ss                                                       ]
-                         ++ [("strToNat",      show s,                   mkThm1 SS.strToNat      strToNat      s      ) | s <- ss                                                       ]
-                         ++ [("natToStr",      show i,                   mkThm1 SS.natToStr      natToStr      i      ) | i <- iUBs                                                     ])
-          ++ map mkTest2 (  [("strToStrAt",    show s, show i,           mkThm2 SS.strToStrAt    strToStrAt    s i    ) | s <- ss, i  <- range s                                        ]
-                         ++ [("strToCharAt",   show s, show i,           mkThm2 SS.strToCharAt   strToCharAt   s i    ) | s <- ss, i  <- range s                                        ]
-                         ++ [("concat",        show s, show s1,          mkThm2 SS.concat        (++)          s s1   ) | s <- ss, s1 <- ss                                             ]
-                         ++ [("isInfixOf",     show s, show s1,          mkThm2 SS.isInfixOf     isInfixOf     s s1   ) | s <- ss, s1 <- ss                                             ]
-                         ++ [("isSuffixOf",    show s, show s1,          mkThm2 SS.isSuffixOf    isSuffixOf    s s1   ) | s <- ss, s1 <- ss                                             ]
-                         ++ [("isPrefixOf",    show s, show s1,          mkThm2 SS.isPrefixOf    isPrefixOf    s s1   ) | s <- ss, s1 <- ss                                             ]
-                         ++ [("take",          show s, show i,           mkThm2 SS.take          genericTake   i s    ) | s <- ss, i <- iUBs                                            ]
-                         ++ [("drop",          show s, show i,           mkThm2 SS.drop          genericDrop   i s    ) | s <- ss, i <- iUBs                                            ]
-                         ++ [("indexOf",       show s, show s1,          mkThm2 SS.indexOf       indexOf       s s1   ) | s <- ss, s1 <- ss                                             ])
-          ++ map mkTest3 (  [("subStr",        show s, show  i, show j,  mkThm3 SS.subStr        subStr        s i  j ) | s <- ss, i  <- range s, j <- range s, i + j <= genericLength s]
-                         ++ [("replace",       show s, show s1, show s2, mkThm3 SS.replace       replace       s s1 s2) | s <- ss, s1 <- ss, s2 <- ss                                   ]
-                         ++ [("offsetIndexOf", show s, show s1, show i,  mkThm3 SS.offsetIndexOf offsetIndexOf s s1 i ) | s <- ss, s1 <- ss, i <- range s                               ])
+genStrings = map mkTest1 (  [("length",        show s,                   mkThm1 SL.length        strLen        s      ) | s <- ss                                                       ]
+                         ++ [("null",          show s,                   mkThm1 SL.null          null          s      ) | s <- ss                                                       ]
+                         ++ [("head",          show s,                   mkThm1 SL.head          head          s      ) | s <- ss, not (null s)                                         ]
+                         ++ [("tail",          show s,                   mkThm1 SL.tail          tail          s      ) | s <- ss, not (null s)                                         ]
+                         ++ [("singleton",     show c,                   mkThm1 SL.singleton     (: [])        c      ) | c <- cs                                                       ]
+                         ++ [("implode",       show s,                   mkThmI SL.implode                     s      ) | s <- ss                                                       ]
+                         ++ [("strToNat",      show s,                   mkThm1 SL.strToNat      strToNat      s      ) | s <- ss                                                       ]
+                         ++ [("natToStr",      show i,                   mkThm1 SL.natToStr      natToStr      i      ) | i <- iUBs                                                     ])
+          ++ map mkTest2 (  [("strToCharAt",   show s, show i,           mkThm2 SL.elemAt        strToCharAt   s i    ) | s <- ss, i  <- range s                                        ]
+                         ++ [("++",            show s, show s1,          mkThm2 (SL.++)          (++)          s s1   ) | s <- ss, s1 <- ss                                             ]
+                         ++ [("isInfixOf",     show s, show s1,          mkThm2 SL.isInfixOf     isInfixOf     s s1   ) | s <- ss, s1 <- ss                                             ]
+                         ++ [("isSuffixOf",    show s, show s1,          mkThm2 SL.isSuffixOf    isSuffixOf    s s1   ) | s <- ss, s1 <- ss                                             ]
+                         ++ [("isPrefixOf",    show s, show s1,          mkThm2 SL.isPrefixOf    isPrefixOf    s s1   ) | s <- ss, s1 <- ss                                             ]
+                         ++ [("take",          show s, show i,           mkThm2 SL.take          genericTake   i s    ) | s <- ss, i <- iUBs                                            ]
+                         ++ [("drop",          show s, show i,           mkThm2 SL.drop          genericDrop   i s    ) | s <- ss, i <- iUBs                                            ]
+                         ++ [("indexOf",       show s, show s1,          mkThm2 SL.indexOf       indexOf       s s1   ) | s <- ss, s1 <- ss                                             ])
+          ++ map mkTest3 (  [("subStr",        show s, show  i, show j,  mkThm3 SL.subList       subStr        s i  j ) | s <- ss, i  <- range s, j <- range s, i + j <= genericLength s]
+                         ++ [("replace",       show s, show s1, show s2, mkThm3 SL.replace       replace       s s1 s2) | s <- ss, s1 <- ss, s2 <- ss                                   ]
+                         ++ [("offsetIndexOf", show s, show s1, show i,  mkThm3 SL.offsetIndexOf offsetIndexOf s s1 i ) | s <- ss, s1 <- ss, i <- range s                               ])
   where strLen :: String -> Integer
         strLen = fromIntegral . length
 
@@ -658,9 +674,6 @@ genStrings = map mkTest1 (  [("length",        show s,                   mkThm1
                                           "" -> -1
                                           (_:r) -> go (i+1) r
 
-        strToStrAt :: String -> Integer -> String
-        s `strToStrAt` i = [s `strToCharAt` i]
-
         strToCharAt :: String -> Integer -> Char
         s `strToCharAt` i = s `genericIndex` i
 
@@ -829,8 +842,8 @@ in8s = xsSigned
 iUBs :: [Integer]
 iUBs = [-1000000] ++ [-1 .. 1] ++ [1000000]
 
-rs :: [AlgReal]
-rs = [fromRational (i % d) | i <- is, d <- dens]
+ars :: [AlgReal]
+ars = [fromRational (i % d) | i <- is, d <- dens]
  where is   = [-1000000] ++ [-1 .. 1] ++ [10000001]
        dens = [5,100,1000000]
 
@@ -874,4 +887,129 @@ se = [Left 3, Right 5]
 st :: [(Integer, Integer)]
 st = [(1, 2), (-1, -5), (0, 9), (5, 5)]
 
+misc :: [TestTree]
+misc = [ testCase "misc-t1" $ assertIsSat t1
+       ]
+ where -- https://stackoverflow.com/questions/69033969/trivial-rationals-problems-without-variables-in-sbv-solver-in-haskell
+       t1 = do _xs <- sRationals []
+               constrain $ (5.%1:: SRational) .<= (5.%1:: SRational)
+
+-- Test these with make test TGT=sEnum_
+genEnums :: [TestTree]
+genEnums =
+    -- Only bounded for from, otherwise infinite (or too big for chars)
+    [mkTest1 "from"       s     (from [s..    ] s) | s <- univ @(WordN 4)]
+ ++ [mkTest1 "from"       s     (from [s..    ] s) | s <- univ @(IntN  4)]
+ ++ [mkTest1 "from"       s     (from [s..    ] s) | s <- univ @Day]
+ ++ [mkTest1 "from"       s     (from [s..    ] s) | s <- w8s]
+ ++ [mkTest1 "from"       s     (from [s..    ] s) | s <- i8s]
+
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- univ @(WordN 4), t <- univ @(WordN 4)]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- univ @(IntN  4), t <- univ @(IntN  4)]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- univ @Day      , t <- univ @Day      ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- w8s            , t <- w8s            ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- i8s            , t <- i8s            ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- ints           , t <- ints           ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- floats         , t <- floats         ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- doubles        , t <- doubles        ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- fps            , t <- fps            ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- lcs            , t <- lcs            ]
+ ++ [mkTest2 "fromTo"     s t   (fromTo [s..t   ] s t) | s <- rs             , t <- rs             ]
+
+    -- Only bounded for fromThen, otherwise infinite (or too big for chars)
+ ++ [mkTest2 "fromThen"   s t   (fromThen [s, t.. ] s t) | s <- univ @(WordN 4), t <- univ @(WordN 4), s /= t]
+ ++ [mkTest2 "fromThen"   s t   (fromThen [s, t.. ] s t) | s <- univ @(IntN  4), t <- univ @(IntN  4), s /= t]
+ ++ [mkTest2 "fromThen"   s t   (fromThen [s, t.. ] s t) | s <- univ @Day      , t <- univ @Day      , s /= t]
+ ++ [mkTest2 "fromThen"   s t   (fromThen [s, t.. ] s t) | s <- w8s            , t <- w8s            , s /= t]
+ ++ [mkTest2 "fromThen"   s t   (fromThen [s, t.. ] s t) | s <- i8s            , t <- i8s            , s /= t]
+
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- univ @(WordN 4), t <- univ @(WordN 4), s /= t, u <- univ @(WordN 4)]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- univ @(IntN  4), t <- univ @(IntN  4), s /= t, u <- univ @(IntN  4)]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- univ @Day      , t <- univ @Day      , s /= t, u <- univ @Day      ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- w8s            , t <- w8s            , s /= t, u <- w8s            ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- i8s            , t <- i8s            , s /= t, u <- i8s            ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- ints           , t <- ints           , s /= t, u <- ints           ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- floats         , t <- floats         , s /= t, u <- floats         ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- doubles        , t <- doubles        , s /= t, u <- doubles        ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- fps            , t <- fps            , s /= t, u <- fps            ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- lcs            , t <- lcs            , s /= t, u <- lcs            ]
+ ++ [mkTest3 "fromThenTo" s t u (fromThenTo [s, t..u] s t u) | s <- rs             , t <- rs             , s /= t, u <- rs             ]
+
+  where mkTest1 pre a     = testCase ("sEnum_" ++ pre ++ "_|" ++ show (kindOf a) ++ "|_" ++ show a)
+        mkTest2 pre a b   = testCase ("sEnum_" ++ pre ++ "_|" ++ show (kindOf a) ++ "|_" ++ show (a, b))
+        mkTest3 pre a b c = testCase ("sEnum_" ++ pre ++ "_|" ++ show (kindOf a) ++ "|_" ++ show (a, b, c))
+
+        from cr a1 = assert $ isTheorem $ do
+                        sa1 <- free_
+                        constrain $ sa1 .== literal a1
+
+                        pure $ [sEnum|sa1..|] .== literal cr
+
+        fromTo cr a1 a2 = assert $ isTheorem $ do
+                            sa1 <- free_
+                            constrain $ sa1 .== literal a1
+
+                            sa2 <- free_
+                            constrain $ sa2 .== literal a2
+
+                            pure $ [sEnum|sa1..sa2|] .== literal cr
+
+        fromThen cr a1 a2 = assert $ isTheorem $ do
+                              sa1 <- free_
+                              constrain $ sa1 .== literal a1
+
+                              sa2 <- free_
+                              constrain $ sa2 .== literal a2
+
+                              pure $ [sEnum|sa1, sa2 ..|] .== literal cr
+
+        fromThenTo cr a1 a2 a3 = assert $ isTheorem $ do
+                                   sa1 <- free_
+                                   constrain $ sa1 .== literal a1
+
+                                   sa2 <- free_
+                                   constrain $ sa2 .== literal a2
+
+                                   sa3 <- free_
+                                   constrain $ sa3 .== literal a3
+
+                                   pure $ [sEnum|sa1, sa2 .. sa3|] .== literal cr
+
+        univ :: (Enum n, Bounded n) => [n]
+        univ = [minBound .. maxBound]
+
+        ints :: [Integer]
+        ints = [-3 .. 3]
+
+        -- Floats create too big a problem for z3, even though we have ground terms. So, skip
+        floats :: [Float]
+        -- floats = [-3.4, -3.2 .. 3.5]
+        floats = []
+
+        -- Ditto here
+        doubles :: [Double]
+        -- doubles = [-3.4, -3.2 .. 3.5]
+        doubles = []
+
+        -- NB. Precision here is important. If you pick too small of a significand
+        -- size then you can turn this enumeration into an infinite list, busting the tests.
+        -- For details see: https://gitlab.haskell.org/ghc/ghc/-/issues/15081
+        -- And likewise, constant folding doesn't happen, too difficult for z3
+        fps :: [FloatingPoint 5 8]
+        -- fps = [-3.4, -3.2 .. 3.5]
+        fps = []
+
+        -- This one works, but is way too slow. So we further reduce the range
+        rs :: [AlgReal]
+        -- rs = [-3.4, -3.2 .. 3.5]
+        rs = [-0.4, -0.2 .. 0.4]
+
+        -- don't add min/max bounds here. causes too big lists.
+        lcs :: [Char]
+        lcs = map C.chr [5, 10, 30, 40, 41, 42, 43, 90, 100]
+
+-- Quiet GHC about unused enum elts
+_unused :: SDay
+_unused = undefined sMon sTue sWed sThu sFri sSat sSun
+
 {- HLint ignore module "Reduce duplication" -}
